//
//  DLRepeatingObject.h
//  DLSubstrate
//
//  Created by Patrick Rogers MC on 13-10-31.
//  Copyright (c) 2013 Marketcircle Inc. All rights reserved.
//

#import <Foundation/Foundation.h>

/**
 * The simplified type combines the type, repeatInterval
 * and other options to identify the repeat types available
 * on the device.
 */
typedef NS_ENUM(NSInteger, GWSimplifiedRepeatType) {
    GWRepeatsNone,
    GWRepeatsDaily,
    GWRepeatsWeekly,
    GWRepeatsEveryTwoWeeks,
    GWRepeatsMonthly,
    GWRepeatsYearly,
    GWCustomRepeat
};

typedef enum {
    GWRepeatFrequencyTypeNone = 1,
    GWRepeatFrequencyTypeDaily = 2,
    GWRepeatFrequencyTypeWeekly = 3,
    GWRepeatFrequencyTypeMonthly = 4,
    GWRepeatFrequencyTypeYearly = 5,
} GWRepeatFrequencyType;

typedef enum {
    DLRepeatingObjectSupportsDuration           = 1L << 0, // The class generates objects that have a duration
    
    DLRepeatingObjectSupportsOccurrenceProxies  = 1L << 1, // The class can generate occurrence proxies based on a repeat rule
    DLRepeatingObjectSupportsOccurrenceHoles    = 1L << 2, // The class can create holes in the repeat chain to suppress certain occurrences
    DLRepeatingObjectSupportsOccurrenceHolePlugs= 1L << 3, // The class can create exception 'plugs' to replace occurrences suppressed by holes
    DLRepeatingObjectSupportsAllFuturePlugs     = 1L << 4, // The class can create a repeating object that breaks the repeat
    // chain and applies to all future occurrences. These objects can
    // also create 'hole' and 'plug' exceptions for the first occurrence in the chain.
    
    DLRepeatingObjectSupportsDAVIdentifier      = 1L << 5, // The class has a DAV identifier
    DLRepeatingObjectSupportsETag               = 1L << 6, // The class has an eTag
    
    DLRepeatingObjectSupportsAll                = 0xFFFFFFFF   // The class supports all defined repeating object features
} DLRepeatingObjectSupportedFeatures;

typedef enum {
    GWRepeatEndTypeEndsNever,
    GWRepeatEndTypeEndsAfterXTimes,
    GWRepeatEndTypeEndsOnDate,
} GWRepeatEndType;

typedef enum {
    DLRepeatingObjectEditingPromptAnswerNotRepeating,    // If the object is not repeating, no question is asked and this is the answer.
    DLRepeatingObjectEditingPromptAnswerAllFuture,       // Edit this and all future occurrences from this instance on
    DLRepeatingObjectEditingPromptAnswerThisOccurrence,  // Edit only this occurrence
    DLRepeatingObjectEditingPromptAnswerCancel           // User hit cancel, so no editing action should be taken.
} DLRepeatingObjectEditingPromptAnswer;

typedef DLRepeatingObjectEditingPromptAnswer (^DLRepeatingObjectPromptCallback)(NSString *detail,NSString *singlePrompt,NSString *allPrompt);

/**
 * DLRepeatingObject is a protocol that should be adopted by persistent DLPObjects that use a repeat
 * sepcification to create non persistent instances (proxies).
 *
 * Classses that conform to DLRepeatingObject must also conform to DLRepeatingObjectPrivate. The
 * reason that DLRepeatingObject doesn't conform to DLRepeatingObjectPrivate is to keep the private
 * methods from being imported publicly in the hopes that people won't try to call them inapropriately.
 */
@protocol DLRepeatingObject <NSObject>
/**
 * For concrete repeating objects repeatParent is the previous object in a repeat chain if the object
 * is an 'all future' exception plug for another repeating object.
 * For proxies repeatParent is the DLRepeatingObject object that the proxy stands in for.
 */
@property(nonatomic,readwrite,strong) DLPObject <DLRepeatingObject> *repeatParentObject;
#pragma mark - EO Properties These properties are defined in the model and implemented by the underlying EO
@property(nonatomic,readwrite,strong) NSDate *repeatOriginalDate;
@property(nonatomic,readwrite,strong) NSNumber *repeatInterval;
@property(nonatomic,readwrite,strong) NSNumber *repeatFrequency;
@property(nonatomic,readwrite,strong) NSDate *repeatEndDate;
@property(nonatomic,readonly,strong) NSDate *rawRepeatEndDate;
@property(nonatomic,readwrite,strong) NSNumber *repeatCount;
@property(nonatomic,readwrite,strong) NSNumber *repeatByNthDay;
@property(nonatomic,readwrite,strong) NSString *repeatByMonths;
@property(nonatomic,readwrite,strong) NSString *repeatByMonthDays;
@property(nonatomic,readwrite,strong) NSString *repeatByDays;
@property(nonatomic,readonly,strong) NSArray *reminders;
@property(nonatomic,readonly,strong) NSDate *modifyDate;

#pragma mark Synthesized Properties -
@property (nonatomic,readonly) BOOL isRepeating;
@property (nonatomic,readonly) BOOL isRepeatOccurrence;
@property (nonatomic,readonly) BOOL isLastRepeatOccurrence;
@property (nonatomic,readonly) BOOL isRepeatByMonthDays;
@property (nonatomic,readonly) BOOL isRepeatByNthDay;
@property (nonatomic,readonly) BOOL isRepeatByDays;
@property (nonatomic,readonly) BOOL isRepeatingOrRepeatOccurrence;
@property (nonatomic,readonly) BOOL processAsFloating;
@property (nonatomic,readonly) NSInteger occurrenceIndex;
@property (nonatomic,readonly) NSInteger nthDayWeekNumber;
@property (nonatomic,readonly) NSInteger nthDayWeekDay;
@property (nonatomic,assign) NSArray *repeatByDaysAsArray;
@property (nonatomic,assign) NSArray *repeatByMonthsAsArray;
@property (nonatomic,assign) NSArray *repeatByMonthDaysAsArray;
@property (nonatomic,readonly) NSSet *repeatExceptionTimes;
@property (nonatomic,readonly) NSString *repeatDescription;
@property (nonatomic,readonly) NSString *shortRepeatDescription;
@property (nonatomic) GWSimplifiedRepeatType simplifiedRepeatType;
@property (nonatomic,retain) NSDictionary *repeatRule;
@property (nonatomic,readonly) NSTimeZone *repeatCalculationTimeZone;
@property (nonatomic,readonly) NSTimeZone *nativeFormatTimeZone;
@property(nonatomic,assign) GWRepeatEndType repeatEndType;
@property(nonatomic,readonly) NSArray *repeatEndTypeEnumWrappers;
@property(nonatomic,assign) MCEnumWrapper *repeatEndTypeEnumWrapper;

/**
 * The time of the root occurrence. Setting the value of the occurrenceStartTime
 * should also update the occurrenceEndTime to preserve the duration of the object.
 */
@property (nonatomic,retain) NSDate *occurrenceStartTime;
@property (nonatomic,readonly) NSDate *occurrenceEndTime;
@property (nonatomic,readonly) NSTimeInterval durationSecondsValue;
/**
 * The raw occurrence start time is only defined for concrete repeating objects, not
 * proxies. Retrieving the rawOccurrenceStartTime on a proxy is undefined since proxies
 * always exist in the local time zone.
 */
@property (nonatomic,readonly) NSDate *rawOccurrenceStartTime;
/**
 * The raw occurrence end time is only defined for concrete repeating objects, not
 * proxies. Retrieving the rawOccurrenceEndTime on a proxy is undefined since proxies
 * always exist in the local time zone.
 */
@property (nonatomic,readonly) NSDate *rawOccurrenceEndTime;

/**
 * Repeating objects that have a dav identifier should expose it here
 * so that the repeat manager can update it when necessary. Objects
 * that do not have a dav identifier may overload this property to be
 * a NOP.
 */
@property(nonatomic,retain) NSString *davIdentifier;

/**
 * Repeating objects that have an eTag should expose it here. The repeat
 * manager itself doesn't use the eTag but some of its tests do.
 */
@property(nonatomic,retain) NSString *eTag;

/**
 * Repeating objects that have an eTag should should implement this
 * method so that the repeat manager can update it when necessary. Objects
 * that do not have an eTag may implement this method to do nothing.
 */
- (void)updateETag;

/**
 * A bit mask of repeating object features that this object supports. Not all repeating
 * objects need to support all features, but they will behave differently if unsupported
 * features are used. See DLRepeatingObjectSupportedFeatures for supported values.
 */
+ (NSUInteger)repeatFeaturesMask;

#pragma mark Repeat Proxy Queries

/**
 * Returns all proxies for this object within a given date range. These methods will only
 * generate repeat proxies for dates that are specified using the repeatRule.
 *
 * @param exceptions A set of exception dates that should not have occurrences even though they
 *                   are specified in the repeat rule.  If exceptions is not specified it will
 *                   be auto generated, but it can be more efficient to pass in the array if the
 *                   caller already has the data.
 */
- (NSDictionary *)repeatProxiesFromTime:(NSDate *)startTime toTime:(NSDate *)endTime;
- (NSDictionary *)repeatProxiesFromTime:(NSDate *)startTime toTime:(NSDate *)endTime exceptionTimes:(NSSet *)exceptions;

- (NSDate *)timeOfNextRepeatOccurrenceAfterTime:(NSDate *)afterTime;
- (NSDate *)timeOfFirstRepeatOccurrence;

#pragma mark Repeat Exception Queries

/**
 * Returns all repeatException times within an interval.
 */
- (NSSet *)repeatExceptionTimesFromTime:(NSDate *)startTime toTime:(NSDate *)endTime;

/**
 * Returns all repeatException objects with an interval.
 */
- (NSArray *)repeatExceptionHolesFromTime:(NSDate *)startTime toTime:(NSDate *)endTime;

/**
 * Returns all repeatException objects linked to self.
 */
- (NSArray *)repeatExceptionHoles;

#pragma mark - Repeat Occurrence Queries

/**
 * Returns TRUE if an occurrence of self occurs on a given date.
 *
 * @param time The date to test.
 * @return TRUE if an occurrence of self occurs at any time on the same
 *         day (as defined by the repeatCalculationTimeZone) as date.
 */
- (BOOL)repeatOccursAtTime:(NSDate *)time;

/**
 * This method determines if there is a repeat occurrence on a specific day.
 * The day is interpreted in the current timezone, meaning that in a different
 * timezone the return value may be different. As a consequence,
 * this method may be used when scheduling an appointment at the users behest,
 * but should not be used when calculating repeat proxies.
 */
- (BOOL)repeatOccursOnDate:(NSDate *)aDate;

- (NSInteger)countValueForRepeatTime:(NSDate *)aDate;

/**
 * This method returns the index number of the occurrence on a specific day.
 * The day is interpreted in the current timezone, meaning that in a different
 * timezone a different occurrence may fall in the bounds of aDate (or no occurrence)
 * and the return value may thus change. As a consequence,
 * this method may be used when scheduling an appointment at the users behest,
 * but should not be used when calculating repeat proxies.
 */
- (NSInteger)countValueForRepeatDate:(NSDate*)aDate;

#pragma mark - Repat Rule Maintenance

- (void)adjustStartTimeForRepeatRule;

#pragma mark - Repeat Rule Specification

- (void)addToRepeatByDays:(NSNumber*)aWeekDay;
- (BOOL)removeFromRepeatByDays:(NSNumber*)aWeekDay;

- (void)addToRepeatByMonths:(NSNumber*)aMonth;
- (BOOL)removeFromRepeatByMonths:(NSNumber*)aMonth;

- (void)addToRepeatByMonthDays:(NSNumber*)aMonthDay;
- (BOOL)removeFromRepeatByMonthDays:(NSNumber*)aMonthDay;

- (BOOL)repeatByDaysContainsWeekDay:(NSInteger)weekDay;
- (BOOL)repeatByMonthDaysContainsMonthDay:(NSInteger)monthDay;

- (void)setRepeatByNthDayWithNthWeek:(NSInteger)weekNum weekDay:(NSInteger)weekDay;
- (BOOL)repeatByNthDayIsEqualToNthWeek:(NSInteger)weekNum weekDay:(NSInteger)weekDay;

- (BOOL)repeatByMonthsContainsMonth:(NSInteger)monthNum;

- (void)clearRepeatRule;

#pragma mark Repeat Description methods

+ (NSString *)descriptionForRepeatType:(const GWSimplifiedRepeatType)RepeatType;
- (NSString *)repeatDescriptionWithEndDescription:(BOOL)includeEndDescription;

/**
 * Determines which object should be selected following a model change that affects objects
 * if self is the currently selected object. This will return the appropriate repeat exception
 * if one was generated from the selected object, return nil if the currently selected object
 * has been deleted or is no longer a selection candidate.
 *
 * @param insertedObjects A set of objects that were inserted in the course of the last
 *                        model update. This set is used to detect which exceptions should
 *                        be selected.
 * @param allObjects A set of objects that are valid candidates for selection.
 *
 * @return The object that should replace this as the selection.
 */
- (DLPObject <DLRepeatingObject> *)updateSelectionAfterChange:(NSSet *)insertedObjects allObjects:(NSSet *)allObjects;

#pragma mark Exceptions

/**
 * duplicateWithoutRepeats creates a copy of an object that has the repeat specification cleared. The copy is not
 * a repeat exception, but it can be used in order to create a repeat exception. There are two ways to generate exceptions.
 * One is to create an exception before editing (which requires prompting the user to see if they want to create the
 * exception for a single instance or all future instances). In this case the exception object is edited directly.
 * The second option is to create a duplicate of an object (or a proxy) and edit the duplicate. When the edits are done
 * the duplicate is passed to createRepeatExceptionOnDate: and will be used to create the exception plug.
 */
- (DLPObject <DLRepeatingObject> *)duplicateWithoutRepeats;

/**
 * When the object is being duplicated for the purposes of creating a repeat exception, this method will be called to determine
 * which relationships to ignore.
 */
- (NSSet  *)relationshipLinksToIgnoreWhenCreatingRepeatException;

/**
 * Creates an exception for a repeating object and performs any necessary modfication to the repeat
 * chain.  This is the only public method that should be used to create a repeat exception and can be sent
 * to either a DLRepeatingObject or a DLRepeatingObjectOccurrenceProxy.
 *
 * If necessary the repeat chain will be modified, and the head of the chain will have the necessary
 * exceptions added to its description to prevent it from recurring at the same time as the exception.
 *
 * If necessary the repeat chain will be modified, and self will have the necessary exceptions added to
 * its description to prevent it from recurring at the same time as the exception.
 *
 *  @param      aDate The date on which to creat the exception. The time component of this date will be set to the start time
 *                    of the object in the objects timezone to fully specify the exception time. If the exception time
 *                    is in a different timezone ensure that this conversion will not alter the date to represent a different day.
 *                    If aDate is equal to the startDate of self then the head of the chain will become the exception, otherwise
 *                    the relevant proxy will be selected and used to create the exception.
 *  @param      allFuture If true the repeat chain is split in two.  The original head of the chain has its repeat count
 *                        limited (by count) to end before aDate and excepObject becomes the head of a new chain with the remaining
 *                        repeats (also limited by count).  If allFuture is false excepObject becomes an exception of object and
 *                        adopts it as its parent.
 *                        In the case where the object occurrence referenced by aDate is the head of the chain and allFuture is
 *                        true no exception is needed, so self (which is the head of the repeat chain) is merged with the exception
 *                        object and the exception timblock does not need to be persisted.
 *                        If aDate references the first occurrence and allFuture is false the repeat chain is searched for the first
 *                        proxy, and it's startDate is assigned to the object at the head of the chain.  The exception becomes the
 *                        new object at the head of the chain, and is adopts the original object as its parent object
 *	@param      excepObject A DLRepeatingObject that defines the exception.  If exceptObject is nil a copy of the original object
 *	                        that occurs on aDate will be created as the exception.  If non nil excepObject should be clone of the
 *							original object. Passing in a non nil excepObject is used when the repeat exception is created after an
 *                          object has been edited by editing a copy of the repeat occurrence. Passing in a nil excepObject is used
 *                          when the repeat exception is created before editing the object.
 *  @return     The object which is the exception. no longer part of the repeat chain, but references object as its parent. It is
 *              now an instance of DLRepeatingObject instead of DLRepeatingObjectOccurrenceProxy.
 *  @see createRepeatExceptionOnDate:originalDate:allFutureRepeats:usingExceptionObject:
 */
- (DLPObject <DLRepeatingObject> *)createRepeatExceptionOnDate:(NSDate *)aDate allFutureRepeats:(BOOL)allFuture usingExceptionObject:(DLPObject <DLRepeatingObject> *)excepObject;

/**
 * Removes an instance of a repeat from a repeating object chain and performs any necessary modfication to the object
 * chain.  If necessary the repeat chain will be modified, and self will have the necessary exceptions added to its description,
 * or the object will be deleted.  Self will be moved to the next available position in the repeat chain, or the entire chain
 * will be deleted if there are no remaining proxies.
 *
 * @param allFutureRepeats If true all future repeats will be removed, otherwise only self will be removed.
 *
 * @return True if the self was deleted, false if the chain was modified.
 */
- (BOOL)deleteRepeatOccurenceAllFutureRepeats:(BOOL)allFuture;

#pragma mark Creating An Editing Occurrence

/**
 * Prompts the user to determine how they want to edit self. If self is a repeating persistent object then the callback
 * will be passed an appropriate query to determine how the object should be edited. The query will take into account
 * what kind of occurrence this is (first, typical, last) to only present the user with valid choices. This method can
 * be used to prompt the user before editing or after editing is complete.
 *
 * @param promptCallback The callback block that will be used to perform the actual query. The promptCallback will be
 *                       called with localized query strings and should return the choice that was selected.
 */
- (DLRepeatingObjectEditingPromptAnswer)promptForEditingObject:(DLRepeatingObjectPromptCallback)promptCallback;

/**
 * Returns an instance for editing a repeating object. It will prompt the user if necessary about what type of exception they would
 * like to generate using the prompt callback and it will create an exception object to edit if necessary or dictated by the prompt
 * answer. If self is a repeat occurrence proxy then an exception will always be generated (since proxies cannot be edited), if self
 * is persistent repeating object then an exception will be generated based on the answer supplied by the user.
 *
 * @param objectContextIn The object context in which to create the editable object.
 * @param prompt The callback block to use to pose the question. See promptForEdigingRepeat: for more details.
 */
- (DLPObject <DLRepeatingObject> *)instanceForEditingInObjectContext:(MCPObjectContext *)objectContextIn queryCallback:(DLRepeatingObjectPromptCallback)prompt;

/**
 * Returns an instance for editing a repeating object based on a query answer that was determined previously. It will create an exception
 * object to edit if necessary or dictated by the prompt answer. If self is a repeat occurrence proxy then an exception will always be
 * generated (since proxies cannot be edited), if self is persistent repeating object then an exception will be generated based on the
 * answer supplied by the user.
 *
 * @param objectContextIn The object context in which to create the editable object.
 * @param answer A previously supplied answer about what type of exception to create for this edit.
 */
- (DLPObject <DLRepeatingObject> *)instanceForEditingInObjectContext:(MCPObjectContext *)objectContextIn promptAnswer:(DLRepeatingObjectEditingPromptAnswer)answer;
@end
