//
//  DLRepeatingObjectOccurrenceProxy.h
//  DLSubstrate
//
//  Created by Patrick Rogers MC on 13-11-22.
//  Copyright (c) 2013 Marketcircle Inc. All rights reserved.
//

#import <Foundation/Foundation.h>

#import "DLPObject.h"

@protocol DLRepeatingObject;

/**
 * DLRepeatingObjectOccurrenceProxy defines a protocol for objects that represent repeat occurrences of
 * objects that conform to the DLRepeatingOjbect protocol. Classes that conform to 
 * DLRepeatingObjectOccurrenceProxy do not need to inherit from NSProxy, but their use is simialar.
 *
 * Every instance of a class that conforms to DLRepeatingObjectOccurrenceProxy references a concrete (persistent)
 * DLRepeatingObject instance. The proxy object has a distinct immutable start and end time, but all other
 * properties and methods are implemented by forwarding them to the concrete instance. The proxy is a 
 * non persistent representation of an occurrence of the persistent object. DLRepeatingObjectOccurrenceProxy does
 * not conform to the DLRepeatingObject protocol because it doesn't implement most of the DLRepeatingObject methods,
 * but instances of DLRepeatingObjectOccurrenceProxy can safely be cast to the type of their repeatParentObject
 * or to a class that conforms to DLRepeatingObject since they will forward all of the methods.
 *
 * For most intents and purposes (other than editing) the proxy can be used as if it were an instance of the
 * persistent class. When a proxy is going to be edited an exception must be created and the edits should
 * be performed on the exception. The exception can either be created using the duplicateWithoutRepeats method
 * and then merged back into the repeat chain in the future when the 'real' exception is generated, or it can
 * be created by calling createRepeatExceptionOnDate: to create the exception before editing takes place.
 * 
 * Since instance of classes that conform to DLRepatingObjectOccurrenceProxy can be used as if they are instances
 * of classes that conform to DLRepeatingObject there is no need for code outside of the repeat manager to import
 * this header and it should be considered private to the repeat system. Likewise any classes that conform to the 
 * protocol are also private. (This is currently not adhered to in some substrate code, and there are some notable
 * exceptions for the purposes of efficiency, but most could be refactored to not require knowledge of 
 * DLRepeatingObjectOccurrenceProxy).
 *
 * Classes that wish to conform to the DLRepeatingObjectOccurrenceProxy protocol may import the code in 
 * DLRepeatingObjectOccurrenceProxyImplementation which defines the basic methods required for implementing
 * an occurrence proxy.
 */
@protocol DLRepeatingObjectOccurrenceProxy <NSObject>
/**
 * For proxies repeatParent is the concrete object that the proxy stands in for. Classes that use 
 * DLRepeatingObjectImplementation.m for the proxy method implementations should define this
 * property to return the value of their parent object (that is defined upon initialization of the
 * proxy object).
 */
@property(nonatomic,readwrite,strong) DLPObject <DLRepeatingObject> *repeatParentObject;
/**
 * The time of this occurrence. The start and end times are constant and set upon initilization.
 * The occurrenceEndTime should be initialized by adding the concrete object's duration to the
 * occurrenctStartTime.
 */
@property (nonatomic,readonly) NSDate *occurrenceStartTime;
@property (nonatomic,readonly) NSDate *occurrenceEndTime;

/**
 * Creates an exception for a proxy and performs any necessary modfication to the primary object
 * chain.
 *
 * If necessary the repeat chain will be modified, and the original object will have the 
 * necessary exceptions added to its description to prevent it from recurring at the same time as the exception.
 *	
 *  @param      allFuture If true the repeat chain is split in two.  The original head of the chain has its repeat count
 *                        limited (by count) to end before aDate and excepObject becomes the head of a new chain with the remaining
 *                        repeats (also limited by count).  If allFuture is false excepObject becomes an exception of selfs
 *                        repeatParentObject and adopts it as its parent.
 *	@param      excepObject An object that defines the exception.  If excepObject is nil a copy of the original object
 *	                           that occurs on aDate will be created as the exception.  excepObject should be clone of the
 *							   original object.
 *  @return     the object which is the exception. no longer part of the repeat chain, but references the parent object.
 */
- (DLPObject <DLRepeatingObject> *)createRepeatExceptionForAllFutureRepeats:(BOOL)allFuture usingExceptionObject:(DLPObject <DLRepeatingObject> *)excepObject;

/**
 * Removes a proxy from a repeating object chain and performs any necessary modfication to the object
 * chain.  If necessary the repeat chain will be modified, and the original object will have the necessary exceptions 
 * added to its description, or the object will be deleted.
 *
 * This method is defined in the DLRepeatingObject protocol as well. Calling it on a DLRepeatingObjectOccurrenceProxy
 * will have a different effect than when called on a DLRepeatingObject in some cases, but it is designed to modify 
 * the repeat chain in the intended way.
 *
 * @param allFutureRepeats If true all future repeats will be removed, otherwise only self will be removed.
 * 
 * @return True if the object that is the head of the chain was deleted, false if the chain was modified.
 */
- (BOOL)deleteRepeatOccurenceAllFutureRepeats:(BOOL)allFuture;
@end
