//
//  DLTimeblockDataSource.h
//  DLApp
//
//  Created by Patrick Rogers on 16/06/08.
//  Copyright 2008 Marketcircle, Inc. All rights reserved.
//

#import <Foundation/Foundation.h>

@class GWUser,GWContact,GWTask,GWTimeblock;
@class DLPObjectContext;
@class BDQualifier;
@class DLTimeblockDataSource;

// Specifies which objects to return when userFilter is nil
typedef enum {
    DLTimeblockDataSourceDefaultUserCurrent,    // Objects owned by the current user (should be deprecated in future, callers can assign it themselves)
    DLTimeblockDataSourceDefaultUserAll,        // Returns all objects (technically the correct behaviour since the dataSource is unfiltered)
    DLTimeblockDataSourceDefaultUserNone        // Returns nothing (usefull in some instances)
} DLTimeblockDataSourceDefaultUser;

typedef enum {
	DLTimeblockDataSourceStartDateDidChange         = 1L<<1,
	DLTimeblockDataSourceEndDateDidChange			= 1L<<2,
	DLTimeblockDataSourceTimeZoneDidChange			= 1L<<3,
	DLTimeblockDataSourceUserFilterDidChange		= 1L<<4,
	DLTimeblockDataSourceResourceFilterDidChange	= 1L<<5,
	DLTimeblockDataSourceAdditionalAndQualifierDidChange = 1L<<6,
	DLTimeblockDataSourceFetchArrayDidChange        = 1L<<7,
	DLTimeblockDataSourceSubscribedCalendarsDidChange   = 1L<<8,
	DLTimeblockDataSourceVisibilityParametersDidChange	= 1L<<9,
	DLTimeblockDataSourceTimeDidChange				= 1L<<10,
	DLTimeblockDataSourceTimeblockDidChange         = 1L<<11,
	DLTimeblockDataSourceTimeblockDidDelete         = 1L<<12,
	DLTimeblockDataSourceTimeblockDidInsert         = 1L<<13,
	DLTimeblockDataSourceRefreshAll                 = 1L<<14,    
	DLTimeblockDataSourceUpdatesDidEnable			= 1L<<15,
} DLTimeblockDataSourceInvalidateReason;

typedef NS_ENUM(NSInteger, DLTimeblockDataSourceBinInterval) {
	DLTimeblockDataSourceBinIntervalNone,
	DLTimeblockDataSourceBinIntervalDay,
	DLTimeblockDataSourceBinIntervalMonth,
	DLTimeblockDataSourceBinIntervalYear,
	DLTimeblockDataSourceBinIntervalLocalizedWeek,
	DLTimeblockDataSourceBinIntervalWeekSunday,
	DLTimeblockDataSourceBinIntervalWeekMonday,
	DLTimeblockDataSourceBinIntervalWeekTuesday,
	DLTimeblockDataSourceBinIntervalWeekWednesday,
	DLTimeblockDataSourceBinIntervalWeekThursday,
	DLTimeblockDataSourceBinIntervalWeekFriday,
	DLTimeblockDataSourceBinIntervalWeekSaturday,
};

extern NSString *DLTimeblockDataSourceDataDidChangeNotification; // User info dictionary will contain an array of modified dates
extern NSString *DLTimeblockDataSourceDataDidChangeUpdatedDates;
extern NSString *DLTimeblockDataSourceDataDidChangeInsertedTimeblocks;
extern NSString *DLTimeblockDataSourceDataDidChangeUpdateStart;
extern NSString *DLTimeblockDataSourceDataDidChangeUpdateEnd;
extern NSString *DLTimeblockDataSourceDataDidChangeIsFlush;
extern NSString *DLTimeblockDataSourceDataDidPrune;
extern NSString *DLTimeblockDataSourceConfigurationDidChangeNotification; // User info dictionary will contain the reason for the invalidation
extern NSString *DLTimeblockDataSourceConfigurationDidChangeReasonKey;
extern NSString *DLTimeblockDataSourceBusyCountDidChange; // The number of busy datasources has changed. This notification may be delivered on a background
                                                          // thread, and the number of busy data sources is guaranteed not to change while it is being processed
extern NSString *DLTimeblockDataSourceBusyCountDidChangeValue; // The number of busy datasources has changed. This notification may be delivered on a background
                                                               // thread, and the number of busy data sources is guaranteed not to change while it is being processed

typedef void (^DLTDSDataDidChangeBlock)(DLTimeblockDataSource *dataSource,NSSet *inserts,NSDate *startDate,NSDate *endDate,NSArray *dates,BOOL isFlush);

/**
 * All ivars can be accessed and updated on the main thread and the main thread only without synchronization unless otherwise noted.
 * Ivars that require synchronization are all synchronized on self.
 */
@interface DLTimeblockDataSource : NSObject {
	NSDate *today;                      /// There will always be a section for this date, even if it has no timeblocks
	
	NSMutableDictionary *datesHash;     /// Storage of arrays of timeblocks, indexed by date. Only updated on the main thread.
	NSMutableArray *datesArray;         /// An ordered array of all dates stored in datesHash. Only updated on the main thread.
    NSMutableSet *allTimeblocks;        /// A set containing all timeblocks stored by self. Only updated on the main thread.
}
@property (nonatomic,weak) DLPObjectContext *dlpObjectContext;
@property (nonatomic,strong) id userFilter;
@property (nonatomic) DLTimeblockDataSourceDefaultUser defaultUserFilter;
@property (nonatomic,readonly) NSArray *userFilterAsArray;
@property (nonatomic,strong) id resourceFilter;
@property (nonatomic,readonly) NSArray *resourceFilterAsArray;
@property (nonatomic,strong) GWContact *contactFilter;
@property (nonatomic,strong) BDQualifier *additionalAndQualifier;
@property (nonatomic,strong) NSDate *startDate; /// The start of the query interval, must be updated or read when synchronized on self
@property (nonatomic,strong) NSDate *endDate;   /// The start of the query interval, must be updated or read when synchronized on self
@property (nonatomic,strong) NSTimeZone *timeZone;      /// The timeZone in which to retrieve floating timeblocks. If nil the default timezone will be used
@property (nonatomic,readonly) NSUInteger numberOfUniqueDates;
@property (nonatomic,readonly) BOOL isBusy;  /// True if the datasource is not suspended and there is data waiting to be fetched or indexed
/**
 * A set of objects to limit the datasource to accesing. Should only be initialized on the main thread before the datasource is active
 * If fetchObjects is non nil only occurrences for tasks and timeblocks that are specified in the array will be returned by the datasource.
 */
@property (nonatomic,strong) NSArray *fetchObjects;
@property (nonatomic,strong) NSArray *subscribedCalendars; /// An array of subscribed calendar ids. All timeblocks that are in these subscriptions and match range and permissions will be returned from this set regardless of the and qualifier
@property (nonatomic,readonly) NSSet *allTimeblocks;
@property (nonatomic,strong) NSSet *timeblockInitialAttributeNames;
@property (nonatomic,strong) NSSet *contactInitialAttributeNames;
@property (nonatomic,strong) NSSet *taskInitialAttributeNames;
@property (nonatomic,strong) NSSet *projectInitialAttributeNames;
@property (nonatomic,strong) NSSet *opportunityInitialAttributeNames;
@property (nonatomic) BOOL ensureTodayIsDefined;
@property (nonatomic) BOOL createProxiesForMultiDayBannerTimeblocks;
@property (nonatomic) BOOL createProxiesForMultiDayNonBannerTimeblocks;
@property (nonatomic) BOOL indexProxies;
@property (nonatomic) BOOL includeRepeatProxies;
@property (nonatomic) BOOL updatesSuspended;
@property (nonatomic) BOOL expandCacheOnOverflow;
@property (nonatomic) NSUInteger maximumTimeblockCount;
@property (nonatomic) NSUInteger maximumRepeatProxyCount;
@property (nonatomic) BOOL includeContactTimeblocks;
@property (nonatomic) BOOL includeTaskTimeblocks;
@property (nonatomic) BOOL includeCancelledMeetings;
@property (nonatomic) BOOL includeProjectTimeblocks;
@property (nonatomic) BOOL includeOpportunityTimeblocks;
@property (nonatomic) BOOL includeNotDoneObjectives;
@property (nonatomic) BOOL includeIncompleteTasks;
@property (nonatomic) BOOL excludePrivateTimeblocks;
@property (nonatomic) BOOL combineOverdueTasksOnToday;
@property (nonatomic) NSTimeInterval allDayDurationThreshold;  // If non zero timeblocks that are longer than this they will be treated as if they were allDay
@property (nonatomic,copy) DLTDSDataDidChangeBlock dataDidChangeBlock;
@property (nonatomic) NSUInteger blockSizeDays;
@property (nonatomic) NSUInteger processingCompleteNotificationInterval;

/**
 * If true the date range will only filter repeating timeblocks, their proxies
 * and their exceptions.  If false it will filter all timeblocks.
 */
@property (nonatomic) BOOL onlyLimitRepeatsByTimeRange;

+ (void)setShowBirthdaysAnniversariesInCalendar:(BOOL)showBirthdayAnniversariesVal;
+ (BOOL)showBirthdaysAnniversariesInCalendar;

+ (void)setShowTasksInCalendar:(BOOL)showTasksVal;
+ (BOOL)showTasksInCalendar;

+ (void)setShowOpportunitiesInCalendar:(BOOL)showVal;
+ (BOOL)showOpportunitiesInCalendar;

+ (void)setShowProjectsInCalendar:(BOOL)showVal;
+ (BOOL)showProjectsInCalendar;

+ (BDQualifier *)myCalendarQualifierForUser:(GWUser *)user;
+ (BDQualifier *)myCalendarQualifierForUser:(GWUser *)user from:(NSDate *)queryStart to:(NSDate *)queryEnd includeCancelled:(BOOL)includeCancelled includeSubscribed:(BOOL)includeSubscribed;
+ (BDQualifier *)timeWhereQualifierFrom:(NSDate *)queryStartDate to:(NSDate *)queryEndDate timeZone:(NSTimeZone *)timeZone onlyLimitRepeatsByTimeRange:(BOOL)onlyLimitRepeatsByTimeRange;

/**
 * Retrieves the global number of DLTimeblockDataSource instances that are in the process
 * of fetching or indexing data and allows the caller to examine the value in a block. If 0
 * all datasources are currently up to date. The number of busy datasources is guaranteed not
 * to be updated while this block is being processed. This method will block until any pending
 * updates to the number of active datasources is complete. This method
 * will block if called while processing a DLTimeblockDataSourceDataDidChangeNotification. In
 * that case you can get the count by accessing the DLTimeblockDataSourceBusyCount in the userInfo
 * of the notification.
 */
+ (void)processActiveDataSourceCount:(void (^)(NSUInteger numberOfBusyDataSources))block;

- (id)initWithStartDate:(NSDate *)startDateInit endDate:(NSDate *)endDateInit objectContext:(DLPObjectContext *)objectContext;

- (void)configureAtomically:(void (^)())configureBlock;

- (void)onlyReturnBusyTimeblocks;
- (void)onlyReturnFreeTimeblocks;

/**
 * Triggers a fetch and cycles the run loop until the query is complete. The results of the current
 * query will have been retrieved and indexed before this method returns.
 */
- (void)fetchSynchronously;

- (NSUInteger)numberOfTimeblocksForNormalizedDate:(NSDate *)queryDate;
- (NSUInteger)numberOfTimeblocksForDate:(NSDate *)queryDate;
- (NSUInteger)numberOfTimeblocksForDateAtIndex:(NSUInteger)dateIndex;

- (NSDate *)dateAtIndex:(NSUInteger)dateIndex;
- (NSDate *)dateAtIndexRelativeToStartDate:(NSUInteger)index;
- (NSDate *)dateForTimeblock:(GWTimeblock *)timeblock;
- (NSInteger)indexForDate:(NSDate *)date;
- (NSInteger)indexForNormalizedDate:(NSDate *)date;
- (NSInteger)indexOfFirstDateBeforeDate:(NSDate *)date;
- (GWTimeblock *)timeblockAtIndex:(NSUInteger)blockIndex forDate:(NSDate *)queryDate;
- (GWTimeblock *)timeblockAtIndex:(NSUInteger)blockIndex forDateAtIndex:(NSUInteger)dateIndex;
- (NSArray *)timeblocksForDate:(NSDate *)queryDate;
- (NSArray *)timeblocksForDateAtIndex:(NSUInteger)dateIndex;
- (NSDictionary *)eventsBinnedByInterval:(DLTimeblockDataSourceBinInterval)binInterval;
- (NSDictionary *)eventsBinnedByDayWithOverlap:(NSTimeInterval)dayExtensionThreshold;

- (void)flushCachedData;

- (NSMutableDictionary *)cacheDictionaryForKey:(NSString *)cacheName;
- (void)releaseCacheForKey:(NSString *)cacheName;
@end
